(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Ground_Context
imports Term
begin

no_notation Term.Hole ("\<box>")

subsection {* Ground Contexts *}

text {* Ground contexts are terms containing arbitrary occurences of holes
instead of variables. *}

(* If 'v is dropped we get a problem with equiv_class below. *)
datatype_new (-:'f,-:'v) gctxt = GCHole ("\<box>")
                        | GCFun 'f "('f, 'v) gctxt list"

fun equiv_class :: "('f, 'v) gctxt \<Rightarrow> ('f, 'v) terms" ("[_]") where
  "[\<box>] = UNIV"
| "[GCFun f Cs] = {Fun f ts | ts. length ts = length Cs \<and> (\<forall>i<length Cs. (ts ! i) \<in> [Cs ! i])}"

text {* An alternative induction scheme for ground contexts. *}
lemma
  fixes P :: "('f, 'v) gctxt \<Rightarrow> bool"
  assumes "P \<box>"
    and "\<And>f Cs. (\<And>C. C \<in> set Cs \<Longrightarrow> P C) \<Longrightarrow> P (GCFun f Cs)"
  shows gctxt_induct[case_names GCHole GCFun, induct type: "gctxt"]: "P C"
  by (rule gctxt.induct, insert assms, auto)

lemma equiv_class_GCFunE:
  assumes "t \<in> [GCFun f Cs]" obtains ts
  where "t = Fun f ts" and "length ts = length Cs" and "\<forall>i<length Cs. (ts ! i) \<in> [Cs ! i]"
  using assms by force

lemma equiv_class_nonempty: "\<exists>t. t \<in> [C]"
proof (induct C)
  case GCHole thus ?case by simp
next
  case (GCFun f Cs)
  hence "\<forall>i<length Cs. \<exists>t. t \<in> [Cs ! i]" by simp
  with bchoice[of "{0..<length Cs}" "\<lambda>i t. t \<in> [Cs ! i]"]
    obtain g where 1: "\<forall>i<length Cs. g i \<in> [Cs ! i]" by (auto simp: all_nat_less_eq)
  let ?ts = "map g [0..<length Cs]"
  have "length ?ts = length Cs" by simp
  moreover have "\<forall>i<length Cs. (?ts ! i) \<in> [Cs ! i]" by (simp add: 1)
  ultimately show ?case unfolding equiv_class.simps by best
qed

lemma equiv_class_GCFun_eq:
  assumes "([GCFun f Cs]::('f, 'v) terms) = [GCFun g Ds]" (is "?T = ?S")
  shows "f = g"
    and "length Cs = length Ds"
    and "\<forall>i<length Cs. ([Cs ! i]::('f, 'v) terms) = [Ds ! i]"
    (is "\<forall>i<length Cs. ?Cs i = ?Ds i")
proof -
  obtain ss and ts where 1: "Fun f ss \<in> ?T" and 2:"Fun g ts \<in> ?S"
    using equiv_class_nonempty[of "GCFun f Cs"]
      and equiv_class_nonempty[of "GCFun g Ds"] by auto
  show f: "f = g" by (rule ccontr) (insert 1 2 assms, auto)
  from equiv_class_GCFunE[OF 1] have len1: "length ss = length Cs"
    and in1: "\<forall>i<length Cs. ss ! i \<in> ?Cs i" by auto
  from equiv_class_GCFunE[OF 2] have len2: "length ts = length Ds"
    and in2: "\<forall>i<length Ds. ts ! i \<in> ?Ds i" by auto
  show len: "length Cs = length Ds"
  proof (rule ccontr)
    assume "length Cs \<noteq> length Ds"
    moreover from equiv_class_GCFunE
      have "\<forall>t \<in> [GCFun g Ds]. num_args t = length Ds" by auto
    ultimately have "Fun g ts \<notin> ?T" using 2 by simp
    with 2 and assms show False by simp
  qed
  show "\<forall>i<length Cs. ?Cs i = ?Ds i"
  proof (rule ccontr)
    assume "\<not> (\<forall>i<length Cs. ?Cs i = ?Ds i)"
    with len obtain i where i: "i < length Cs" and ne: "?Cs i \<noteq> ?Ds i" by auto
    then obtain t where "(t \<in> ?Cs i \<and> t \<notin> ?Ds i) \<or> (t \<notin> ?Cs i \<and> t \<in> ?Ds i)" by blast
    thus False
    proof (elim disjE conjE)
      assume "t \<in> ?Cs i" and "t \<notin> ?Ds i"
      hence notin: "\<forall>ts. length ts = length Ds \<and> ts ! i = t \<longrightarrow> ts ! i \<notin> ?Ds i" by simp
      from equiv_class_nonempty have "\<forall>i<length Cs. \<exists>t. t \<in> ?Cs i" by force
      from Ex_list_of_length_P[OF this]
        obtain ts where 1: "length ts = length Cs" and 2: "\<forall>i<length Cs. (ts ! i) \<in> ?Cs i"
        by auto
      let ?ts = "ts[i := t]"
      from 1 have 3: "length ?ts = length Cs" by simp
      have 4: "\<forall>i<length Cs. (?ts ! i) \<in> ?Cs i"
      proof (intro allI impI)
        fix j assume j: "j < length Cs"
        show "?ts ! j \<in> ?Cs j"
          using i 1 2 j `t \<in> ?Cs i` by (cases "i = j") simp_all
      qed
      from 3 and 4 have inT: "Fun f ?ts \<in> ?T" by simp
      from notin[THEN spec[of _ "?ts"]] and 3[unfolded len] and i[unfolded len]
        have "?ts ! i \<notin> ?Ds i" by simp
      hence "Fun f ?ts \<notin> ?S" using i[unfolded len] by (auto simp: f)
      with inT and assms show False by blast
    next
      assume "t \<notin> ?Cs i" and "t \<in> ?Ds i"
      hence notin: "\<forall>ss. length ss = length Cs \<and> ss ! i = t \<longrightarrow> ss ! i \<notin> ?Cs i" by simp
      from equiv_class_nonempty have "\<forall>i<length Ds. \<exists>t. t \<in> ?Ds i" by force
      from Ex_list_of_length_P[OF this]
        obtain ss where 1: "length ss = length Ds" and 2: "\<forall>i<length Ds. (ss ! i) \<in> ?Ds i"
        by auto
      let ?ss = "ss[i := t]"
      from 1 have 3: "length ?ss = length Ds" by simp
      have 4: "\<forall>i<length Ds. (?ss ! i) \<in> ?Ds i"
      proof (intro allI impI)
        fix j assume j: "j < length Ds"
        show "?ss ! j \<in> ?Ds j"
          using i 1 2 j `t \<in> ?Ds i` by (cases "i = j") simp_all
      qed
      from 3 and 4 have inS: "Fun g ?ss \<in> ?S" by simp
      from notin[THEN spec[of _ "?ss"]] and 3[unfolded len[symmetric]] and i
        have "?ss ! i \<notin> ?Cs i" by simp
      hence "Fun g ?ss \<notin> ?T" using i[unfolded len] by (auto simp: f len)
      with inS and assms show False by blast
    qed
  qed
qed

text {*
Equality of two equiv_classes is the same as syntactic equality of ground contexts.
*}
lemma equiv_class_eq_eq_conv:
  shows "([C]::('f, 'v) terms) = [D] \<longleftrightarrow> C = D"
    (is "?T = ?S \<longleftrightarrow> _")
proof
  assume "C = D" thus "?T = ?S" by simp
next
  assume "?T = ?S"
  thus "C = D"
  proof (induct C arbitrary: D)
    case GCHole thus ?case by (cases D) (force)+
  next
    case (GCFun f Cs)
    note IH = this
    thus ?case
    proof (induct D)
      case GCHole thus ?case by force
    next
      case (GCFun g Ds)
      let ?Cs = "\<lambda>i. [Cs ! i]::('f, 'v) terms"
      let ?Ds = "\<lambda>i. [Ds ! i]::('f, 'v) terms"     
      from equiv_class_GCFun_eq[OF GCFun(3)]
        have f: "f = g" and len: "length Cs = length Ds" and "\<forall>i<length Cs. ?Cs i = ?Ds i"
        by simp_all
      from GCFun(2) and this(3) have "\<forall>i<length Cs. Cs ! i = Ds ! i" by force
      with len have "Cs = Ds" by (rule nth_equalityI)
      with f show ?case by simp
    qed
  qed
qed

lemma equiv_class_mono:
  assumes "[s] \<subseteq> [t]"
  shows "[GCFun f (bef @ s # aft)] \<subseteq> [GCFun f (bef @ t # aft)]"
    (is "[GCFun f ?bsa] \<subseteq> [GCFun f ?bta]")
proof -
  {
    fix i u
    assume i: "i < Suc (length bef + length aft)" (is "i < ?n")
    assume elem: "u \<in> [?bsa ! i]"
    from i have "i < length bef \<or> i = length bef \<or> (\<exists> k. i = Suc (length bef) + k \<and> k < length aft)" (is "?one \<or> ?two \<or> ?three") by arith
    hence "u \<in> [?bta ! i]" 
    proof 
      assume ?one
      hence "?bta ! i = ?bsa ! i" by (auto simp: nth_append)
      thus ?thesis using elem by auto 
    next
      assume "?two \<or> ?three"
      thus ?thesis
      proof
	assume ?two
	hence "?bsa ! i = s \<and> ?bta ! i = t" by (auto simp: nth_append)
	with assms elem show ?thesis by auto
      next
	assume ?three
	from this obtain k where "i = Suc (length bef) + k \<and> k < length aft" by auto
	hence "?bta ! i = ?bsa ! i" by (auto simp: nth_append)
	with elem show ?thesis by auto	
      qed
    qed
  }
  thus ?thesis by auto
qed


subsection {* Ground Context Matching *}

definition match :: "('f, 'v) gctxt \<Rightarrow> ('f, 'v) term \<Rightarrow> bool" where
  "match h t = (\<exists>\<sigma>. t \<cdot> \<sigma> \<in> [h])"


(* reestablish "default" setup *)
no_notation Ground_Context.GCHole ("\<box>")
no_notation Ground_Context.equiv_class ("[_]")
notation Term.Hole ("\<box>")
hide_const (open) match

end
